#!/usr/bin/perl
#
# updateProjectFiles.pl  written by Gregor Rebel 2011
#
# Automatically generates list of all source-code and header files being used 
# by current makefile-project.
#
# + How it works
#   (1) runs "make all -p" to retrieve all path variables from make
#   (2) processed main.c + all included files (depth-search)
#   (3) writes out all processed files
#

my $ProjectFile = shift(@ARGV);
my @Files2Process = @ARGV;

unless ($ProjectFile) {
  print "$0 PROJECT_FILE [A.C] [B.C]..\n";
  print "  PROJECT_FILE  filename of QtCreator .files file (will be filled with all found files)\n";
  print "  A.C,B.C,..    start points of recursive search\n";
  print "\n";
  print "  $0 will run make all -p to retrieve all include- and source-paths being known to the compiler.\n";
  print "  It will then recursively scan all given .c-files for included files. All found files will be \n";
  print "  added to PROJECT_FILE.\n";
  print "\n";
  print "  Example: $0 TemplateName.files\n";
  exit 5;
}

my $Makefile = `make all -p 2>&1`;
$Makefile =~ s/\n\r/\n/g;
$Makefile =~ s/\r\n/\n/g;
$Makefile =~ s/\\\n//g;
my @ErrorMessages;

#{ extract list of c-source search paths (vpath %.c)       -> @VPathC, %VPathC
my $Pos = index($Makefile, 'vpath %.c ') + 10;
my $EolPos = findEndOfLine($Makefile, $Pos);
my @NewVPathC = split(":", substr($Makefile, $Pos, $EolPos - $Pos + 1));
@NewVPathC = map { $_ =~ s/\r//g; $_ =~ s/\n//g; $_; }  @NewVPathC;
my @VPathC; my %VPathC;
print "VPathC: ".join(',', addEntries(\@VPathC, \%VPathC, @NewVPathC))."\n";
#}
#{ extract list of assembler code search paths (vpath %.s) -> @VPathS, %VPathS
my $Pos = index($Makefile, 'vpath %.s ') + 10;
my $EolPos = findEndOfLine($Makefile, $Pos);
my @NewVPathS = split(":", substr($Makefile, $Pos, $EolPos - $Pos + 1));
@NewVPathS = map { $_ =~ s/\r//g; $_ =~ s/\n//g; $_; }  @NewVPathS;
my @VPathS; my %VPathS;
print "VPathS: ".join(',', addEntries(\@VPathS, \%VPathS, @NewVPathS))."\n";
#}
#{ extract list of common search paths (VPATH)             -> @VPath,  %VPath
my $Pos = index($Makefile, 'Variable »VPATH«)');
my $EolPos = findEndOfLine($Makefile, $Pos);
$Pos = index($Makefile, '# ', $EolPos) + 2; # pos @ after # in next line
$EolPos = findEndOfLine($Makefile, $Pos);
my @NewVPath = split(":", substr($Makefile, $Pos, $EolPos - $Pos + 1));
@NewVPath = map { $_ =~ s/\r//g; $_ =~ s/\n//g; $_; }  @NewVPath;

my @VPath; my %VPath;
addEntries(\@VPath, \%VPath, '.');
print "VPath: ".join(',', addEntries(\@VPath, \%VPath, @NewVPath))."\n";
#}
#{ extract list of include paths                           -> @INCLUDE_DIRS, %INCLUDE_DIRS
my @INCLUDE_DIRS; my %INCLUDE_DIRS;
my $Pos = index($Makefile, 'INCLUDE_DIRS =');
if ($Pos > -1) {
  $Pos += 14;
  my $EolPos = findEndOfLine($Makefile, $Pos);
  my @IncludeDirs = split( " -I ", substr($Makefile, $Pos, $EolPos - $Pos) );
  @IncludeDirs = map { $_ =~ s/\r//g; $_ =~ s/\n//g; $_; }  @IncludeDirs;
  unless ($IncludeDirs[0]) { shift(@IncludeDirs); }
  foreach my $IncludeDir (@IncludeDirs) {
    my %RemoveChar = (' ' => 1, '/' => 1);
    while ( $RemoveChar{substr($IncludeDir, -1)}) { substr($IncludeDir, -1) = undef; }
  }
  addEntries(\@INCLUDE_DIRS, \%INCLUDE_DIRS,  @IncludeDirs);
  print "INCLUDE_DIRS: ".join(',', @INCLUDE_DIRS)."\n";
}
#}
#{ extract list of main objects                            -> @MainObjects, %MainObjects
my @MainObjects; my %MainObjects;
my $Pos = index($Makefile, 'main.elf: ');
if ($Pos > -1) {
  $Pos += 10;
  my $EolPos = findEndOfLine($Makefile, $Pos);
  my @Objects = split( " ", substr($Makefile, $Pos, $EolPos - $Pos) );
  print "MainObjects: ".join(',', addEntries(\@MainObjects, \%MainObjects,  @Objects))."\n";
}
unless (@Files2Process)  # no source-files given: use main-objects  
  { @Files2Process = map { substr($_, 0, -1).'c'; } @MainObjects; }

#}

my @FilesProcessed;
my %FilesProcessed;
my @FilesFound;
my %FilesFound;
my %FilesNotFound;

my %ProjectFiles;
if (0) { # read in existing entries from exiting .files
  if (-e $ProjectFile) {
    open(IN, "<$ProjectFile") or die("$0 ERROR: Cannot read from project file '$ProjectFile' ($!)\n");
    my @Lines = <IN>;
    close(IN);
    foreach my $File (@Lines) {
      $ProjectFiles{strip($File)} = 1;
    }
  }
}

processFiles(0, '', @Files2Process);

$ProjectFiles{makefile} = 1;
open(OUT, ">$ProjectFile") or die("$0 ERROR: Cannot write to project file '$ProjectFile' ($!)\n");
print OUT join("\n", sort keys %ProjectFiles)."\n";
close(OUT);
exit 0;

sub processFiles() {            # scan given files for includes
  my $RecursionDepth = shift;   # increased by every recursion
  my $IncludedBy     = shift;   # name of file which included @Files2Process
  my @Files2Process = @_;
  my $Spaces;
  for (my $I = 0; $I < $RecursionDepth; $I++) { $Spaces .= '  '; }
  
  foreach my $File2Process (@Files2Process) {        # scan for all included files
    @ErrorMessages = ();
    my $FilePath = findFile($File2Process);
    unless ($FilePath) {
      if ($RecursionDepth > 0) 
        { print $Spaces."2+ $IncludedBy #include $File2Process - ERROR: ".shift(@ErrorMessages)."!\n"; }
      else
        { print "$File2Process - ERROR: ".shift(@ErrorMessages)."!\n"; }
    }
    else {
      if ( addEntries(\@FilesProcessed, \%FilesProcessed, $FilePath) ) {
        if (open(IN, "<$FilePath")) {
          $ProjectFiles{ $FilePath } = 1;
          if ($RecursionDepth > 0)
            { print $Spaces."3+ $IncludedBy: #include $File2Process ($FilePath)\n"; }
          else
            { print "1+ $File2Process ($FilePath)\n"; }
          my @Lines = <IN>;
          close(IN);
          my $Lines = join("", @Lines);
          if (0) { # concatenate multi-lines (corrupts line-number calculation)
            $Lines =~ s/\n\r/\n/g;
            $Lines =~ s/\r\n/\n/g;
            $Lines =~ s/\\\n//g;
          }
          
          my $Pos = index($Lines, '#include ');
          while ($Pos > 0) {
            $Pos += 9;
            my $StartPos = findNext($Lines, $Pos, '<', '"') + 1;
            my $EndPos;
            my $StartChar = substr($Lines, $Pos, 1);
            if ($StartChar eq '<') {
              $EndPos = findNext($Lines, $Pos + 1, '>') - 1;
            }
            if ($StartChar eq '"') {
              $EndPos = findNext($Lines, $Pos + 1, '"') - 1;
            }
            if ($EndPos > $StartPos) { # found include filename
              my $IncludeName = substr($Lines, $StartPos, $EndPos - $StartPos + 1);
              my $LineNo = getLineNo($Lines, $StartPos);

              if ( addEntries(\@FilesFound, \%FilesFound, $IncludeName) ) {
                # print $Spaces."| new include '$IncludeName'\n";
                processFiles($RecursionDepth+1, $File2Process.':'.$LineNo, $IncludeName);
              }
            }
          
            $Pos = index($Lines, '#include ', $Pos);
          }
        }
        else { print "$0 - ERROR: Cannot read from file '$FilePath'"; }
      }
    }
  }
  if (0) { print "DEBUG: \n'".join("'\n'", keys %ProjectFiles)."\n"; }
}

my %fF_Folders;
my %fF_FilePaths;
sub findFile() {               # finds relative or absolute file-path for given filenames
  my @FileNames = @_;
  
  my @FilePaths;
  
  foreach my $FileName (@FileNames) {
    unless (defined $fF_FilePaths{$FileName}) { # file-name new: search in all corresponding directories
      my @CorrespondingDirectories = @VPath;
      my $Suffix = getSuffix($FileName);
      if ($Suffix eq 's') { push @CorrespondingDirectories, @VPathS; }
      if ($Suffix eq 'c') { push @CorrespondingDirectories, @VPathC; }
      if ($Suffix eq 'h') { push @CorrespondingDirectories, @INCLUDE_DIRS; }
      
      foreach $CorrespondingDirectory (@CorrespondingDirectories) {
        my $FilePath = $CorrespondingDirectory.'/'.$FileName;
        
        if ( -e  $FilePath) { $fF_FilePaths{$FileName} = $FilePath; break; }
      }
      unless ($fF_FilePaths{$FileName}) { # ERROR
        unless ($FilesNotFound{$FileName}) {
          push @ErrorMessages, "Cannot find $FileName (.$Suffix) in any of ".
                               join(' ', @CorrespondingDirectories)."\n";
        }
        $FilesNotFound{$FileName} = 1;
      }
      #X else { $ProjectFiles{ $fF_FilePaths{$FileName} } = 1; }
    }
    push @FilePaths, $fF_FilePaths{$FileName};
  }
  
  if (wantarray) { return @FilePaths; }
  else           { return $FilePaths[0]; }
}
sub findNext() {               # finds next occurence of one of given list of characters
  my $Text       = shift;
  my $Position   = shift;
  my @Chars2Find = @_;
  
  my $Length = length($Text);
  my %Chars2Find;
  foreach my $Char2Find (@Chars2Find)
  { $Chars2Find{$Char2Find} = 1; }
  
  while ( ( ! $Chars2Find{ substr($Text, $Position , 1) } ) &&
          ($Position < $Length)
        )
        { $Position++; }
  
  return $Position;
}
sub findStartOfLine() {        # returns position of first  non-white character in current line
  my $Text     = shift;
  my $Position = shift;
  my $C;
  
  do {
    $C = substr($Text,  $Position, 1);
    if ( ($C ne "\n") && ($C ne "\r") ) { $Position--; }
  } while ( ($Position > 0) && ($C ne "\n") && ($C ne "\r") );
  my %WhiteChar = ( " " => 1, "\t" => 1 );
  
  while ( $WhiteChar{substr($Text,  $Position, 1)} && ($Position > 1) ) 
  { $Position++; }             # remove leading spaces
  
  return $Position;
}
sub findEndOfLine() {          # returns position of last non-white character before end of line
  my $Text     = shift;
  my $Position = shift;
  my $C;
  
  my $Length = length($Text);
  do {
    $C = substr($Text,  $Position, 1);
    if ( ($C ne "\n") && ($C ne "\r") ) { $Position++; }
  } while ( ($Position < $Length) && ($C ne "\n") && ($C ne "\r") );
  my %WhiteChar = ( " " => 1, "\t" => 1 );
  
  while ( $WhiteChar{substr($Text,  $Position, 1)} && ($Position > 1) ) 
  { $Position--; }             # remove trailing spaces
  
  return $Position;
}
sub addEntries() {             # adds given strings to list-ref + hash-ref if new
  my $ListRef = shift;
  my $HashRef = shift;
  
  my @AddedEntries;
  while (@_) {
    my $Entry   = shift(@_);
    
    unless ($HashRef->{$Entry}) {
      $HashRef->{$Entry} = 1;
      push @$ListRef, $Entry;
      push @AddedEntries, $Entry;
    }
  }
  
  if (wantarray) { return @AddedEntries; }
  else           { return $AddedEntries[0]; }
}
sub getSuffix() {              # returns filename suffix
  my $FileName = shift();
  
  my $LastDotPos = rindex($FileName, '.');
  if ($LastDotPos > -1) { return lc( substr($FileName, $LastDotPos + 1) ); }
  return undef;
}
sub getLineNo() {              # returns line number of given byte position in given string
  my $Text     = shift;
  my $Position = shift;
  
  $Text = substr($Text, 0, $Position + 1);
  
  my $LineNo = 0;
  my $LfPos = index($Text, "\n");
  while ($LfPos > -1) {
    $LineNo++;
    $LfPos = index($Text, "\n", $LfPos + 1);
  }
  
  return $LineNo;
}
my %s_WhiteChar;
sub strip() {                  # removes trailing and leading whitechars from given string
  my $String = shift;
  
  unless (%s_WhiteChar) {
    $s_WhiteChar{"\n"} = 1;
    $s_WhiteChar{"\r"} = 1;
    $s_WhiteChar{"\t"} = 1;
    $s_WhiteChar{" "}  = 1;
  }
  while ( $s_WhiteChar{substr($String, 0, 1)} ) { substr($String, 0, 1) = undef; }
  while ( $s_WhiteChar{substr($String, -1, 1)} )   { substr($String, -1)   = undef; }
  
  return $String;
}

exit 0;
